#ifndef _ASM_M32R_NAMEI_H
#define _ASM_M32R_NAMEI_H

/* $Id$ */

/* orig : i386 2.4.18 */

/*
 * linux/include/asm-m32r/namei.h
 *
 * Included from linux/fs/namei.c
 */

/* This dummy routine maybe changed to something useful
 * for /usr/gnemul/ emulation stuff.
 * Look at asm-sparc/namei.h for details.
 */

#define __emul_prefix() NULL

#endif /* _ASM_M32R_NAMEI_H */
