#ifndef __LCD_LPH8923_H
#define __LCD_LPH8923_H

enum lcd_lph8923_test_num {
	LCD_LPH8923_TEST_RGB_LINES,
};

enum lcd_lph8923_test_result {
	LCD_LPH8923_TEST_SUCCESS,
	LCD_LPH8923_TEST_INVALID,
	LCD_LPH8923_TEST_FAILED,
};

#endif
