/*
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 *
 * Copyright (C) 2006 by Ralf Baechle (ralf@linux-mips.org)
 */
#ifndef _ASM_DS1742_H
#define _ASM_DS1742_H

#include <ds1742.h>

#endif /* _ASM_DS1742_H */
